#include "deformobj.h"

void DeformObj::
initialize_block_mesh(const Vec2f &centre, float length, float mesh_size)
{
   // first handle the nodes
   int nx=(int)ceil(length/mesh_size)+1;
   n=nx*nx;
   restx.resize(n);
   surface_nodes.resize(4*nx-4);
   v.resize(n);
   f.resize(n);
   mass.resize(n);
   int i, j, k=0, s=0;
   for(i=0; i<nx; ++i) for(j=0; j<nx; ++j){
      restx[k]=centre+length*Vec2f(i/(float)nx-0.5, j/(float)nx-0.5);
      if(i==0 || i==nx-1 || j==0 || j==nx-1){
	 surface_nodes[s]=k;
	 ++s;
      }
      ++k;
   }
   x=restx;
   oldx=restx;
   // now handle the triangles
   nt=2*(nx-1)*(nx-1);
   tri.resize(nt);
   stiffness.resize(nt);
   int t=0;
   for(i=1; i<nx; ++i) for(j=1; j<nx; ++j){
      tri[t]=Vec3i(i-1+nx*(j-1), i+nx*(j-1), i+nx*j); ++t;
      tri[t]=Vec3i(i-1+nx*(j-1), i+nx*j, i-1+nx*j); ++t;
   }
   // and record the mesh size
   min_mesh_length=length/(nx-1);
}

void DeformObj::
compute_mass(float density_)
{
   density=density_;
   for(int i=0; i<n; ++i)
      mass[i]=0;
   for(int t=0; t<nt; ++t){
      int i=tri[t][0], j=tri[t][1], k=tri[t][2];
      float contribution=density*mag(cross(restx[i]-restx[k], restx[j]-restx[k]))/6;
      mass[i]+=contribution;
      mass[j]+=contribution;
      mass[k]+=contribution;
   }
}

void DeformObj::
compute_stiffness(float youngs_modulus_, float poisson_ratio_)
{
   youngs_modulus=youngs_modulus_;
   poisson_ratio=poisson_ratio_;
   for(int t=0; t<nt; ++t)
      stiffness[t].compute(youngs_modulus, poisson_ratio, restx[tri[t][0]], restx[tri[t][1]], restx[tri[t][2]]);
}

void DeformObj::
set_damping(float damping_)
{
   damping=damping_;
   for(int t=0; t<nt; ++t)
      stiffness[t].damping_ratio=damping;
}

void DeformObj::
advance(float dt)
{
   compute_forces();
   // just Symplectic Euler for now
   for(int i=0; i<n; ++i){
      v[i]+=(dt/mass[i])*f[i];
      x[i]+=v[i];
   }
}

void DeformObj::
compute_forces(void)
{
   for(int i=0; i<n; ++i)
      f[i]=Vec2f(0.f);
   external_forces.add_forces(mass, x, v, f);
   for(int t=0; t<nt; ++t)
      add_triangle_forces(t);
}

void DeformObj::
add_triangle_forces(int t)
{
   int i=tri[t][0], j=tri[t][1], k=tri[t][2];

   // figure out how the element has rotated
   float cosine, sine;
   polar_rotation(restx[i], restx[j], restx[k], x[i], x[j], x[k], cosine, sine);

   // compute the forces on the corners of the element
   Vec2f f1, f2, f3;
   // Rotate the positions back so we are left with just pure deformation, then subtract off the rest positions
   // to get the displacements. Feed these into the element force computation to get forces.
   stiffness[t].apply_elastic(rotate(cosine, -sine, x[i])-restx[i],
                              rotate(cosine, -sine, x[j])-restx[j],
                              rotate(cosine, -sine, x[k])-restx[k],
                              f1, f2, f3);
   // Rotate those forces back to the current configuration, add to our global force vector
   f[i]+=rotate(cosine, sine, f1);
   f[j]+=rotate(cosine, sine, f2);
   f[k]+=rotate(cosine, sine, f3);
   // same kind of thing for damping
   stiffness[t].apply_damping(rotate(cosine, -sine, v[i]),
                              rotate(cosine, -sine, v[j]),
                              rotate(cosine, -sine, v[k]),
                              f1, f2, f3);
   f[i]+=rotate(cosine, sine, f1);
   f[j]+=rotate(cosine, sine, f2);
   f[k]+=rotate(cosine, sine, f3);
}

void DeformObj::
compute_max_timestep(void)
{
   // From youngs_modulus, poisson_ratio, damping, density, and min_mesh_length
   // get an estimate for the maximum stable Symplectic Euler time step.
   // Remember to multiply by a safety factor like 0.5!
   maxdt=@@@;
}

