#include <cstdio>
#include <iostream>
#include <fstream>
#include "gluvi.h"
#include "simulation.h"

using namespace std;

Simulation sim;
float dt;
float density;
float youngs_modulus, poisson_ratio, damping;

void init_simulation()
{
   // feel free to alter the initial conditions

   dt=0.01;
   density=1;
   youngs_modulus=10;
   poisson_ratio=0.3;
   damping=0.5;

   sim.object.resize(1);
   sim.object[0].initialize_block_mesh(Vec2f(0.5, 0.5), 0.3, 0.05);
   sim.object[0].compute_mass(density);
   sim.object[0].compute_stiffness(youngs_modulus, poisson_ratio);
   sim.object[0].set_damping(damping);
   sim.object[0].compute_max_timestep();

   sim.collisions.boxleft=0;
   sim.collisions.boxright=1;
   sim.collisions.boxfloor=0;
   sim.collisions.boxceiling=1;
   sim.collisions.boxfriction=0.3;
   sim.collisions.objectfriction=0.1;
}

void advance_simulation(int ignored)
{
   sim.advance(dt);
   glutPostRedisplay();
   glutTimerFunc(50, advance_simulation, 0);
}

void display(void)
{
   glDisable(GL_LIGHTING);

   // draw the box the simulation is supposed to stay inside
   glColor3f(1, 1, 1);
   glBegin(GL_LINE_STRIP);
   glVertex2f(sim.collisions.boxleft, sim.collisions.boxfloor);
   glVertex2f(sim.collisions.boxright, sim.collisions.boxfloor);
   glVertex2f(sim.collisions.boxright, sim.collisions.boxceiling);
   glVertex2f(sim.collisions.boxleft, sim.collisions.boxceiling);
   glVertex2f(sim.collisions.boxleft, sim.collisions.boxfloor);
   glEnd();

   // draw all the elements
   glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
   glBegin(GL_TRIANGLES);
   for(unsigned int i=0; i<sim.object.size(); ++i){
      glColor3f(i/(float)sim.object.size(), 1-i/(float)sim.object.size(), 0.5+0.5*(i%2));
      for(int t=0; t<sim.object[i].nt; ++t){
	 int a=sim.object[i].tri[t][0], b=sim.object[i].tri[t][1], c=sim.object[i].tri[t][2];
	 glVertex2fv(sim.object[i].x[a].v);
	 glVertex2fv(sim.object[i].x[b].v);
	 glVertex2fv(sim.object[i].x[c].v);
      }
   }
   glEnd();
}

struct TimeStepSlider : public Gluvi::Slider
{
   float mindt, maxdt;
   TimeStepSlider(float mindt_, float maxdt_)
      : Gluvi::Slider("timestep", 200, (int)(200*(dt-mindt_)/(maxdt_-mindt_)), 100),
        mindt(mindt_), maxdt(maxdt_)
   {}

   void action()
   {
      dt=(position/(float)length)*(maxdt-mindt) + mindt;
   }
};

struct DensitySlider : public Gluvi::Slider
{
   float mindensity, maxdensity;
   DensitySlider(float mindensity_, float maxdensity_)
      : Gluvi::Slider("density", 200, (int)(200*(density-mindensity_)/(maxdensity_-mindensity_)), 100),
        mindensity(mindensity_), maxdensity(maxdensity_)
   {}

   void action()
   {
      density=(position/(float)length)*(maxdensity-mindensity) + mindensity;
      for(unsigned int i=0; i<sim.object.size(); ++i){
         sim.object[i].compute_mass(density);
	 sim.object[i].compute_max_timestep();
      }
   }
};

struct MaterialSlider : public Gluvi::Slider
{
   float *parameter;
   float minparam, maxparam;
   MaterialSlider(const char *text_, float *parameter_, float minparam_, float maxparam_)
      : Gluvi::Slider(text_, 200, (int)(200*(*parameter_-minparam_)/(maxparam_-minparam_)), 100),
        parameter(parameter_), minparam(minparam_), maxparam(maxparam_)
   {}

   void action()
   {
      *parameter=(position/(float)length)*(maxparam-minparam) + minparam;
      for(unsigned int i=0; i<sim.object.size(); ++i){
         sim.object[i].compute_stiffness(youngs_modulus, poisson_ratio);
	 sim.object[i].compute_max_timestep();
      }
   }
};

struct DampingSlider : public Gluvi::Slider
{
   float mindamping, maxdamping;
   DampingSlider(float mindamping_, float maxdamping_)
      : Gluvi::Slider("damping", 200, (int)(200*(damping-mindamping_)/(maxdamping_-mindamping_)), 100),
        mindamping(mindamping_), maxdamping(maxdamping_)
   {}

   void action()
   {
      damping=(position/(float)length)*(maxdamping-mindamping) + mindamping;
      for(unsigned int i=0; i<sim.object.size(); ++i){
         sim.object[i].set_damping(damping);
	 sim.object[i].compute_max_timestep();
      }
   }
};

int main(int argc, char **argv)
{
   Gluvi::init("deformable2d", &argc, argv);

   glClearColor(0, 0, 0, 0);

   init_simulation();
   glutTimerFunc(50, advance_simulation, 0);

   Gluvi::PanZoom2D cam(-0.1, -0.1, 1.5);
   Gluvi::camera=&cam;

   Gluvi::userDisplayFunc=display;

   // still need to add mouse forces...

   TimeStepSlider dtslider(0.005, 0.1);
   Gluvi::root.list.push_back(&dtslider);

   DensitySlider rhoslider(1, 10);
   Gluvi::root.list.push_back(&rhoslider);

   MaterialSlider ymslider("Young's modulus", &youngs_modulus, 10, 200),
                  prslider("Poisson ratio", &poisson_ratio, 0, 0.49);
   Gluvi::root.list.push_back(&ymslider);
   Gluvi::root.list.push_back(&prslider);

   DampingSlider dampingslider(0, 2);
   Gluvi::root.list.push_back(&dampingslider);
  
   Gluvi::run();
   return 0;
}

